from typing import Any, Dict, List, Optional, Text

import rasax.community.config as rasa_x_config
from rasax.community.database.service import DbService
from rasax.community.database.workspace import Workspace
from sanic.request import Request


class WorkspacesError(Exception):
    """Base exception for all Workspaces errors."""


class WorkspacesNameExists(WorkspacesError):
    """Exception that is raised when there's already a workspace with a given name."""


class WorkspacesIdNotExists(WorkspacesError):
    """Exception that is raised when there's no workspace with a give ID."""


class WorkspacesHasDefaultName(WorkspacesError):
    """Exception that is raised when user tries to rename a default workspace."""


class WorkspacesDraftsCannotReject(WorkspacesError):
    """Exception that is raised when drafts cannot be rejected."""


class WorkspacesService(DbService):
    """Service for all Workspaces related SQL operations."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "WorkspacesService":
        """Creates a `WorkspacesService` from an incoming HTTP request's DB session.

        Args:
            request: Incoming HTTP request.
            **kwargs: other key-value args, not used.

        Returns:
            `WorkspacesService` instance with a connection to the DB.
        """
        return WorkspacesService(request.ctx.db_session)

    def create_workspace(self, name: Text) -> Dict[Text, Any]:
        """Creates a workspace with a given name.

        Args:
            name: Name to assign to the new workspace.

        Raises:
            ValueError: If the workspace name is already taken.

        Returns:
            New workspace.
        """
        workspace = self.query(Workspace).filter(Workspace.name == name).first()
        if workspace:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        workspace = Workspace(name=name)
        self.add(workspace)
        self.flush()

        return workspace.as_dict()

    def get_workspaces(self) -> List[Dict[Text, Any]]:
        """Retrieves a list of existing workspaces.

        Returns:
           Existing workspaces.
        """
        return [workspace.as_dict() for workspace in self.query(Workspace)]

    def update_workspace(self, workspace_id: int, name: Text) -> Dict[Text, Any]:
        """Updates workspace name by id.

        Args:
            workspace_id: ID of the workspace to be updated.
            name: New name to assign to the workspace.

        Raises:
            ValueError: If workspace cannot be found by id, the name is taken,
                or the workspace is a default workspace.

        Returns:
            Updated workspace.
        """
        # check if there is already another workspace with the same name
        workspace_same_name = (
            self.query(Workspace).filter(Workspace.name == name).first()
        )
        if workspace_same_name:
            raise WorkspacesNameExists(f"Workspace with name '{name}' already exists.")

        # check that workspace to be renamed exists
        workspace = self.query(Workspace).filter(Workspace.id == workspace_id).first()
        if not workspace:
            raise WorkspacesIdNotExists(f"Workspace with id {workspace_id} not found.")

        # check if this is a user's default workspace
        if self.is_a_default_workspace(workspace_id):
            raise WorkspacesHasDefaultName(
                f"Workspace with id {workspace_id} is a default workspace."
            )

        workspace.name = name
        return workspace.as_dict()

    def is_a_default_workspace(self, workspace_id: int) -> bool:
        """Checks if workspace is any user's default workspace.

        Args:
            workspace_id: ID of workspace that potentially is a default workspace.

        Returns:
            `True` in case the workspace is a default workspace for some user.
        """
        from rasax.community.services.user_service import UserService

        user_service = UserService(self.session)
        users = user_service.fetch_all_users(rasa_x_config.team_name)
        for user in users:
            if user["default_workspace_id"] == int(workspace_id):
                return True
        return False

    def get_workspace_by_id(self, id: int) -> Optional[Dict[Text, Any]]:
        """Retrieves existing workspace by ID.

        Args:
            id: ID of the workspace.

        Returns:
            List of existing workspaces.
        """
        workspace = self.query(Workspace).filter(Workspace.id == id).first()
        return workspace.as_dict() if workspace else None

    def reject_drafts_from_workspace(self, draft_ids: List[int]) -> List[int]:
        """Rejects drafts from a workspace.

        Args:
            draft_ids: IDs of the drafts to be rejected.

        Returns:
            IDs of rejected drafts.
        """
        from rasax.community.services.logs_service import LogsService

        result = []
        logs_service = LogsService(self.session)
        try:
            for draft_id in draft_ids:
                result.append(
                    logs_service.restore_original_log_state(draft_id).get("id")
                )
        except ValueError as e:
            raise WorkspacesDraftsCannotReject(f"Failed to reject drafts, error: {e}")

        return result
